import pathlib
from uuid import uuid4

import numpy as np
from Bio import SeqIO

from primalscheme3.core.classes import FKmer, PrimerPair, RKmer
from primalscheme3.core.config import IUPAC_ALL_ALLOWED_DNA, Config
from primalscheme3.core.digestion import digest, generate_valid_primerpairs
from primalscheme3.core.errors import (
    MSAFileInvalid,
    MSAFileInvalidBase,
    MSAFileInvalidLength,
)
from primalscheme3.core.mapping import create_mapping, ref_index_to_msa
from primalscheme3.core.seq_functions import remove_end_insertion


def parse_msa(msa_path: pathlib.Path) -> tuple[np.ndarray, dict]:
    """
    Parses a multiple sequence alignment (MSA) file in FASTA format.

    This function reads an MSA file, validates its format and content, and returns a numpy array of the sequences
    and a dictionary with additional information. It checks for sequences of different lengths, empty columns,
    and non-DNA characters. It also removes end insertions from the sequences.

    Args:
        msa_path (pathlib.Path): The path to the MSA file to be parsed.

    Returns:
        tuple: A tuple containing two elements:
            - np.ndarray: A 2D numpy array where each row represents a sequence in the MSA and each column represents a position in the alignment.
            - dict: A dictionary with additional information about the MSA (currently not implemented, returns an empty dict).

    Raises:
        MSAFileInvalidLength: If the MSA contains sequences of different lengths.
        MSAFileInvalid: If the MSA file is empty or not in FASTA format.
        ValueError: If the MSA contains empty columns.
        MSAFileInvalidBase: If the MSA contains non-DNA characters.
    """
    try:
        records_index = SeqIO.index(
            str(msa_path),
            "fasta",
        )
    except ValueError as e:
        raise MSAFileInvalid(f"{msa_path.name}: {e}") from e

    try:
        array = np.array(
            [record.seq.upper() for record in records_index.values()],
            dtype="U1",
            ndmin=2,  # Enforce 2D array even if one genome
        )
    except ValueError as e:
        raise MSAFileInvalidLength(
            f"MSA ({msa_path.name}): contains sequences of different lengths"
        ) from e

    # Check for empty MSA, caused by no records being parsed
    if array.size == 0:
        raise MSAFileInvalid(
            f"No sequences in MSA ({msa_path.name}). Please ensure the MSA uses .fasta format."
        )

    empty_set = {"", "-"}

    empty_col_indexes = []
    # Check for empty columns and non DNA characters
    for col_index in range(0, array.shape[1]):
        slice: set[str] = set(array[:, col_index])
        # Check for empty columns
        if slice.issubset(empty_set):
            empty_col_indexes.append(col_index)
        # Check for non DNA characters
        if slice.difference(IUPAC_ALL_ALLOWED_DNA):
            base_str = ", ".join(slice.difference(IUPAC_ALL_ALLOWED_DNA))
            raise MSAFileInvalidBase(
                f"MSA ({msa_path.name}) contains non DNA characters ({base_str}) at column: {col_index}"
            )
    # Remove empty columns
    array = np.delete(array, empty_col_indexes, axis=1)

    # Remove end insertions
    array = remove_end_insertion(array)

    return array, dict(records_index)


class MSA:
    # Provided
    name: str
    path: str
    msa_index: int

    # Calculated on init
    array: np.ndarray
    _uuid: str
    _chrom_name: str  # only used in the primer.bed file and html report
    _mapping_array: np.ndarray
    _ref_to_msa: dict[int, int]
    _seq_dict: dict

    # Calculated on evaluation
    fkmers: list[FKmer]
    rkmers: list[RKmer]
    primerpairs: list[PrimerPair]

    def __init__(
        self,
        name: str,
        path: pathlib.Path,
        msa_index: int,
        mapping: str,
        progress_manager,
        logger=None,
    ) -> None:
        self.name = name
        self.path = str(path)
        self.msa_index = msa_index
        self.logger = logger
        self.progress_manager = progress_manager

        # Add empty lists for the primerpairs, fkmers and rkmers
        self.primerpairs = []
        self.fkmers = []
        self.rkmers = []

        # Read in the MSA
        try:
            self.array, self._seq_dict = parse_msa(path)
        except Exception as e:
            # Log the error and raise it
            if self.logger:
                self.logger.error(f"MSA: {self.name} failed QC: {e}")
            raise e

        # Create the mapping array
        # Goes from msa idx -> ref idx
        if mapping == "consensus":
            self._chrom_name = self.name + "_consensus"
            self._mapping_array = np.array([*range(len(self.array[0]))])
        elif mapping == "first":
            self._mapping_array, self.array = create_mapping(self.array, 0)
            self._chrom_name = list(self._seq_dict)[0]
        else:
            raise ValueError(f"Mapping method: {mapping} not recognised")

        # Goes from ref idx -> msa idx
        self._ref_to_msa = ref_index_to_msa(self._mapping_array)

        # Assign a UUID
        self._uuid = str(uuid4())[:8]

        if "/" in self._chrom_name:
            new_chromname = self._chrom_name.replace("/", "_")
            warning_str = (
                f"Replacing '/' with '-'. '{self._chrom_name}' -> '{new_chromname}'"
            )
            if self.logger:
                self.logger.warning(warning_str)
            else:
                print(warning_str)
            self._chrom_name = new_chromname

        # Check length
        if len(self._chrom_name) > 200:  # limit is 255
            new_chromname = self._chrom_name[:200]
            if self.logger:
                self.logger.warning(
                    f"Chromname '{self._chrom_name}' is too long, "
                    f"limit is 100 characters. Truncating to '{new_chromname}'"
                )
            self._chrom_name = new_chromname

    def digest(
        self,
        config: Config,
        indexes: tuple[list[int], list[int]] | None = None,
    ) -> None:
        """
        Digest the given MSA array and return the FKmers and RKmers.

        :param cfg: A dictionary containing configuration parameters.
        :param indexes: A tuple of MSA indexes for (FKmers, RKmers), or False to use all indexes.
        :return: None (Class is updated inplace)
        """
        # Create all the kmers
        self.fkmers, self.rkmers = digest(
            msa_array=self.array,
            config=config,
            indexes=indexes,
            logger=self.logger,
            progress_manager=self.progress_manager,
            chrom=self.name,
        )
        # remap the fkmer and rkmers if needed
        if self._mapping_array is not None:
            mapping_set = set(self._mapping_array)

            remaped_fkmers = [fkmer.remap(self._mapping_array) for fkmer in self.fkmers]  # type: ignore
            self.fkmers = [
                x
                for x in remaped_fkmers
                if x is not None and x.end in mapping_set and min(x.starts()) >= 0
            ]
            remaped_rkmers = [rkmer.remap(self._mapping_array) for rkmer in self.rkmers]  # type: ignore
            self.rkmers = [
                x
                for x in remaped_rkmers
                if x is not None
                and x.start in mapping_set
                and max(x.ends()) < self.array.shape[1]
            ]

    def generate_primerpairs(
        self, amplicon_size_min: int, amplicon_size_max: int, dimerscore: float
    ) -> None:
        self.primerpairs = generate_valid_primerpairs(
            fkmers=self.fkmers,
            rkmers=self.rkmers,
            amplicon_size_min=amplicon_size_min,
            amplicon_size_max=amplicon_size_max,
            dimerscore=dimerscore,
            msa_index=self.msa_index,
            progress_manager=self.progress_manager,
            chrom=self.name,
        )
        # Update primerpairs to include the chrom_name and amplicon_prefix
        for primerpair in self.primerpairs:
            primerpair.chrom_name = self._chrom_name
            primerpair.amplicon_prefix = self._uuid

    def write_msa_to_file(self, path: pathlib.Path):
        # Writes the msa to file with parsed chrom names
        with open(path, "w") as outfile:
            for r in self._seq_dict.values():
                # Format each record
                r.id = "_".join(r.id.split("/"))
                r.description = ""
                r.seq = r.seq.upper()
                # Write
                record_str = r.format("fasta")
                outfile.write(record_str)



# Modules
from collections import Counter
from enum import Enum
from typing import Callable, Union

import numpy as np

# Submodules
from primaldimer_py import (
    do_pools_interact_py,  # type: ignore
    which_kmers_pools_interact,  # type: ignore
)

from primalscheme3.core.classes import FKmer, PrimerPair, RKmer
from primalscheme3.core.config import Config
from primalscheme3.core.errors import (
    ERROR_SET,
    ContainsInvalidBase,
    CustomErrors,
    CustomRecursionError,
    GapOnSetBase,
    WalksOut,
    WalksTooFar,
)
from primalscheme3.core.get_window import get_r_window_FAST2
from primalscheme3.core.progress_tracker import ProgressManager
from primalscheme3.core.seq_functions import (
    expand_ambs,
    get_most_common_base,
    reverse_complement,
)
from primalscheme3.core.thermo import (
    THERMORESULT,
    calc_tm,
    thermo_check_kmers,
)


class DIGESTION_ERROR(Enum):
    """
    Enum for the different types of errors that can occur during digestion
    """

    WALKS_OUT = "WalksOut"
    CONTAINS_INVALID_BASE = "ContainsInvalidBase"
    CUSTOM_RECURSION_ERROR = "CustomRecursionError"
    CUSTOM_ERRORS = "CustomErrors"
    GAP_ON_SET_BASE = "GapOnSetBase"
    HAIRPIN_FAIL = "HairpinFail"
    DIMER_FAIL = "DimerFail"  # Interaction within the kmer
    WALK_TO_FAR = "WalkToFar"  # When indels causes the walk to go to far
    AMB_FAIL = "AmbFail"  # Generic error for when the error is unknown

    # Thermo errors
    THERMO_HIGH_GC = "HighGC"
    THERMO_LOW_GC = "LowGC"
    THERMO_HIGH_TM = "HighTM"
    THERMO_LOW_TM = "LowTM"
    THERMO_MAX_HOMOPOLY = "MaxHomopoly"
    THERMO_HAIRPIN = "Hairpin"
    NO_SEQUENCES = "NoSequences"


def parse_error(results: set[CustomErrors | str]) -> DIGESTION_ERROR:
    """
    Parses the error set for the error that occurred
    As only one error is returned, there is an arbitrary hierarchy of errors
    - CONTAINS_INVALID_BASE > GAP_ON_SET_BASE > WALKS_OUT > CUSTOM_RECURSION_ERROR > WALK_TO_FAR > CUSTOM_ERRORS
    """
    if ContainsInvalidBase() in results:
        return DIGESTION_ERROR.CONTAINS_INVALID_BASE
    elif GapOnSetBase() in results:
        return DIGESTION_ERROR.GAP_ON_SET_BASE
    elif WalksOut() in results:
        return DIGESTION_ERROR.WALKS_OUT
    elif CustomRecursionError() in results:
        return DIGESTION_ERROR.CUSTOM_RECURSION_ERROR
    elif CustomErrors() in results:
        return DIGESTION_ERROR.CUSTOM_ERRORS
    elif WalksTooFar() in results:
        return DIGESTION_ERROR.WALK_TO_FAR
    else:  # Return a generic error
        return DIGESTION_ERROR.AMB_FAIL


def parse_thermo_error(result: THERMORESULT) -> DIGESTION_ERROR:
    """
    Parses the THERMORESULT for the error that occurred
    """
    match result:
        case THERMORESULT.HIGH_GC:
            return DIGESTION_ERROR.THERMO_HIGH_GC
        case THERMORESULT.LOW_GC:
            return DIGESTION_ERROR.THERMO_LOW_GC
        case THERMORESULT.HIGH_TM:
            return DIGESTION_ERROR.THERMO_HIGH_TM
        case THERMORESULT.LOW_TM:
            return DIGESTION_ERROR.THERMO_LOW_TM
        case THERMORESULT.MAX_HOMOPOLY:
            return DIGESTION_ERROR.THERMO_MAX_HOMOPOLY
        case THERMORESULT.HAIRPIN:
            return DIGESTION_ERROR.THERMO_HAIRPIN
        case _:
            raise ValueError("Unknown error occurred")


def parse_error_list(
    error_list: list[str | CustomErrors],
) -> list[str | DIGESTION_ERROR]:
    """
    Parses a list of errors and returns a list of DIGESTION_ERROR
    """
    return_list = []
    for result in error_list:
        if isinstance(result, str):
            return_list.append(result)
        elif isinstance(result, CustomErrors):
            return_list.append(parse_error({result}))
    return return_list


def generate_valid_primerpairs(
    fkmers: list[FKmer],
    rkmers: list[RKmer],
    amplicon_size_min: int,
    amplicon_size_max: int,
    dimerscore: float,
    msa_index: int,
    progress_manager: ProgressManager,
    chrom: str = "",
) -> list[PrimerPair]:
    """Generates valid primer pairs for a given set of forward and reverse kmers.

    Args:
        fkmers: A list of forward kmers.
        rkmers: A list of reverse kmers.
        cfg: A dictionary containing configuration parameters.
        msa_index: An integer representing the index of the multiple sequence alignment.
        disable_progress_bar: A boolean indicating whether to disable the progress bar.

    Returns:
        A list of valid primer pairs.
    """
    ## Generate all primerpairs without checking
    checked_pp = []
    pt = progress_manager.create_sub_progress(
        iter=fkmers, process="Generating primer pairs", chrom=chrom
    )
    for fkmer in pt:
        fkmer_start = min(fkmer.starts())
        # Get all rkmers that would make a valid amplicon
        pos_rkmer = get_r_window_FAST2(
            kmers=rkmers,
            start=fkmer_start + amplicon_size_min,
            end=fkmer_start + amplicon_size_max,
        )
        for rkmer in pos_rkmer:
            # Check for interactions
            if not which_kmers_pools_interact(
                [fkmer], [rkmer], dimerscore, calc_all=False
            ):
                checked_pp.append(PrimerPair(fkmer, rkmer, msa_index))

        # Update the count
        pt.manual_update(count=len(checked_pp))

    checked_pp.sort(key=lambda pp: (pp.fprimer.end, -pp.rprimer.start))
    return checked_pp


def walk_right(
    array: np.ndarray,
    col_index_right: int,
    col_index_left: int,
    row_index: int,
    seq_str: str,
    config: Config,
) -> Union[set[str], Exception]:
    """
    Walks to the right of the array and returns a set of valid sequences.

    Args:
        array: A numpy array of DNA sequences.
        col_index_right: The current column index to the right.
        col_index_left: The current column index to the left.
        row_index: The current row index.
        seq_str: The current sequence string.
        config: The configuration object.

    Returns:
        A set of valid DNA sequences or an exception if an error occurs.

    Raises:
        WalksOut: If the function walks out of the array size.
        ContainsInvalidBase: If the sequence contains an invalid base.
    """
    # Guard for correct tm
    if (
        calc_tm(
            seq_str,
            mv_conc=config.mv_conc,
            dv_conc=config.dv_conc,
            dna_conc=config.dna_conc,
            dntp_conc=config.dntp_conc,
        )
        >= config.primer_tm_min
    ):
        return {seq_str}

    # Guard prevents walking out of array size
    if col_index_right >= array.shape[1] - 1 or col_index_left >= array.shape[1] - 1:
        raise WalksOut()

    # Guard for walking too far
    if col_index_right - col_index_left >= config.primer_max_walk:
        raise WalksTooFar()

    new_base = array[row_index, col_index_right]

    # Fix incomplete ends
    if new_base == "":
        new_base = get_most_common_base(array, col_index_right + 1)
    new_string = (seq_str + new_base).replace("-", "")

    # Prevent Ns from being added
    if "N" in new_string:
        raise ContainsInvalidBase()

    # Guard for invalid bases in the sequence
    exp_new_string: set[str] | None = expand_ambs([new_string])
    if exp_new_string is None:
        raise ContainsInvalidBase()

    passing_str = []
    for exp_str in exp_new_string:
        results = wrap_walk(
            walk_right,
            array,
            col_index_right + 1,
            col_index_left,
            row_index,
            exp_str,
            config,
        )
        passing_str.extend(results)

    return passing_str  # type: ignore


def walk_left(
    array: np.ndarray,
    col_index_right: int,
    col_index_left: int,
    row_index: int,
    seq_str: str,
    config: Config,
) -> set[str] | Exception:
    """
    Recursively walks left from a given starting position in a 2D numpy array of DNA bases,
    constructing a set of valid DNA sequences that meet certain criteria.

    Args:
        array: A 2D numpy array of DNA bases.
        col_index_right: The rightmost column index of the region of interest.
        col_index_left: The current leftmost column index of the region of interest.
        row_index: The current row index of the region of interest.
        seq_str: The current DNA sequence being constructed.
        cfg: A dictionary of configuration parameters.

    Returns:
        A set of valid DNA sequences that meet the criteria specified in the function body.

    Raises:
        WalksOut: If the function attempts to walk out of the array.
        ContainsInvalidBase: If the constructed sequence contains an invalid DNA base.
    """

    # Guard prevents walking out of array size
    if col_index_left <= 0 or col_index_right <= 0:
        raise WalksOut()

    # Guard for correct tm
    if (
        calc_tm(
            seq_str,
            mv_conc=config.mv_conc,
            dv_conc=config.dv_conc,
            dna_conc=config.dna_conc,
            dntp_conc=config.dntp_conc,
        )
        >= config.primer_tm_min
    ):
        return {seq_str}

    # Guard for walking too far
    if col_index_right - col_index_left >= config.primer_max_walk:
        raise WalksTooFar()

    new_base = array[row_index, col_index_left - 1]

    # Ensure it can repair truncated regions
    if new_base == "":
        new_base = get_most_common_base(array, col_index_left - 1)
    new_string = (new_base + seq_str).replace("-", "")

    # Guard prevents seqs with an N
    if "N" in new_string:
        raise ContainsInvalidBase()

    # If invalid bases return None
    exp_new_string: set[str] | None = expand_ambs([new_string])
    if exp_new_string is None:
        raise ContainsInvalidBase()

    passing_str = []
    for exp_str in exp_new_string:
        results = wrap_walk(
            walk_left,
            array=array,
            col_index_right=col_index_right,
            col_index_left=col_index_left - 1,
            row_index=row_index,
            seq_str=exp_str,
            config=config,
        )
        passing_str.extend(results)

    return passing_str  # type: ignore


def wrap_walk(
    walkfunction: Callable,
    array: np.ndarray,
    col_index_right: int,
    col_index_left: int,
    row_index: int,
    seq_str: str,
    config: Config,
) -> list[str | CustomErrors]:
    return_list = []
    try:
        seqs = walkfunction(
            array=array,
            col_index_right=col_index_right,
            col_index_left=col_index_left,
            row_index=row_index,
            seq_str=seq_str,
            config=config,
        )
    except CustomErrors as e:
        return_list.append(e)
    except Exception as e:
        raise e
    else:
        return_list.extend(seqs)

    return return_list


def r_digest_to_count(
    align_array: np.ndarray, config: Config, start_col: int, min_freq: float
) -> tuple[int, dict[str | DIGESTION_ERROR, int]]:
    """
    Returns the count of each sequence / error at a given index
    A value of -1 in the return dict means the function returned early, and not all seqs were counted. Only used for WALKS_OUT and GAP_ON_SET_BASE
    """

    ### Process early return conditions
    # If the initial slice is outside the range of the array
    if start_col + config.primer_size_min >= align_array.shape[1]:
        return (start_col, {DIGESTION_ERROR.WALKS_OUT: -1})

    # Check for gap frequency on first base
    base, counts = np.unique(align_array[:, start_col], return_counts=True)
    first_base_counter = dict(zip(base, counts))
    first_base_counter.pop("", None)

    num_seqs = np.sum(counts)
    first_base_freq = {k: v / num_seqs for k, v in first_base_counter.items()}

    # If the freq of gap is above minfreq
    if first_base_freq.get("-", 0) > min_freq:
        return (start_col, {DIGESTION_ERROR.GAP_ON_SET_BASE: -1})

    ### Calculate the total number of sequences
    # Create a counter
    total_col_seqs: Counter[str | DIGESTION_ERROR] = Counter()
    for row_index in range(0, align_array.shape[0]):
        # Check if this row starts on a gap, and if so update the counter and skip
        if align_array[row_index, start_col] == "-":
            total_col_seqs.update([DIGESTION_ERROR.GAP_ON_SET_BASE])
            continue

        start_array = align_array[
            row_index, start_col : start_col + config.primer_size_min
        ]
        start_seq = "".join(start_array).replace("-", "")

        # Prevent Ns from being added
        if "N" in start_seq:
            total_col_seqs.update([DIGESTION_ERROR.CONTAINS_INVALID_BASE])
            continue

        if not start_seq:  # If the start seq is empty go to the next row
            continue

        # Get all sequences
        results = wrap_walk(
            walk_right,
            array=align_array,
            col_index_right=start_col + config.primer_size_min,
            col_index_left=start_col,
            row_index=row_index,
            seq_str=start_seq,
            config=config,
        )
        # If all mutations matter, return on any Error
        if min_freq == 0 and set(results) & ERROR_SET:
            return (start_col, {parse_error(set(results)): -1})

        # Add the results to the Counter
        total_col_seqs.update(
            {seq: 1 / len(results) for seq in parse_error_list(results)}
        )

    return (start_col, dict(total_col_seqs))


def process_seqs(
    seq_counts: dict[str | DIGESTION_ERROR, int], min_freq, ignore_n: bool = False
) -> DIGESTION_ERROR | dict[str, float]:
    """
    Takes the output from *_digest_to_count and returns a set of valid sequences. Or the error that occurred.

    Args:
        col (int): The column number.
        seq_counts (dict[str | DIGESTION_ERROR, int]): A dictionary containing sequence counts.
        min_freq: The minimum frequency threshold.

    Returns:
        DIGESTION_ERROR | dict[str, float]: either an error or a dictionary of parsed sequences.
    """
    # Check for early return conditions
    for error, count in seq_counts.items():
        if count == -1 and isinstance(error, DIGESTION_ERROR):
            return error

    # Remove Ns if asked
    if ignore_n:
        seq_counts.pop(DIGESTION_ERROR.CONTAINS_INVALID_BASE, None)

    total_values = sum(seq_counts.values())
    # Filter out values below the threshold freq
    above_freq_seqs: dict[str | DIGESTION_ERROR, float] = {
        k: v / total_values
        for k, v in seq_counts.items()
        if v / total_values > min_freq
    }
    parsed_seqs: dict[str, float] = {}
    # Guard: If wanted_seqs contains errors return None
    for seq in above_freq_seqs.keys():
        if isinstance(seq, DIGESTION_ERROR):
            return seq
        elif isinstance(seq, str):
            parsed_seqs[seq] = above_freq_seqs[seq]

    return parsed_seqs


def r_digest_index(
    align_array: np.ndarray, config: Config, start_col: int, min_freq: float
) -> RKmer | tuple[int, DIGESTION_ERROR]:
    """
    This will try and create a RKmer started at the given index
    :align_array: The alignment array
    :config: The configuration object
    :start_col: The column index to start the RKmer
    :min_freq: The minimum frequency threshold

    :return: A RKmer object or a tuple of (start_col, error)
    """
    # Count how many times each sequence / error occurs
    _start_col, seq_counts = r_digest_to_count(align_array, config, start_col, min_freq)
    tmp_parsed_seqs = process_seqs(seq_counts, min_freq, ignore_n=config.ignore_n)
    if isinstance(tmp_parsed_seqs, DIGESTION_ERROR):
        return (start_col, tmp_parsed_seqs)
    elif isinstance(tmp_parsed_seqs, dict):
        parsed_seqs = tmp_parsed_seqs
    else:
        raise ValueError("Unknown error occurred")

    if not parsed_seqs:
        return (start_col, DIGESTION_ERROR.NO_SEQUENCES)

    # Create the Kmer
    rc_seqs = [reverse_complement(seq) for seq in parsed_seqs.keys()]
    tmp_kmer = RKmer(start_col, rc_seqs)

    # Thermo check the kmers
    thermo_result = thermo_check_kmers(tmp_kmer.seqs, config)
    match thermo_result:
        case THERMORESULT.PASS:
            pass
        case _:
            return (start_col, parse_thermo_error(thermo_result))

    # Check for dimer
    if do_pools_interact_py([*tmp_kmer.seqs], [*tmp_kmer.seqs], config.dimer_score):
        return (start_col, DIGESTION_ERROR.DIMER_FAIL)
    # All checks pass return the kmer
    return tmp_kmer


def f_digest_to_count(
    align_array: np.ndarray, config: Config, end_col: int, min_freq: float
) -> tuple[int, dict[str | DIGESTION_ERROR, int]]:
    """
    This will try and create a FKmer ended at the given index
    :return: A FKmer object or a tuple of (end_col, error)
    """

    # Check for gap frequency on first base
    base, counts = np.unique(
        align_array[:, end_col], return_counts=True
    )  # -1 for non-inclusive end
    first_base_counter = dict(zip(base, counts))
    first_base_counter.pop("", None)

    num_seqs = np.sum(counts)

    first_base_freq = {k: v / num_seqs for k, v in first_base_counter.items()}

    # If the freq of gap is above minfreq
    if first_base_freq.get("-", 0) > min_freq:
        return (end_col, {DIGESTION_ERROR.GAP_ON_SET_BASE: -1})

    # If the initial slice is outside the range of the array
    if end_col - config.primer_size_min < 0:
        return (end_col, {DIGESTION_ERROR.WALKS_OUT: -1})

    total_col_seqs: Counter[str | DIGESTION_ERROR] = Counter()
    for row_index in range(0, align_array.shape[0]):
        # Check if this row starts on a gap, and if so update the counter and skip
        if align_array[row_index, end_col] == "-":
            total_col_seqs.update([DIGESTION_ERROR.GAP_ON_SET_BASE])
            # Skip to next row
            continue

        start_seq = "".join(
            align_array[row_index, end_col - config.primer_size_min : end_col]
        ).replace("-", "")

        # Prevent Ns from being added
        if "N" in start_seq:
            total_col_seqs.update([DIGESTION_ERROR.CONTAINS_INVALID_BASE])
            continue

        if not start_seq:  # If the start seq is empty go to the next row
            continue

        results = wrap_walk(
            walk_left,
            array=align_array,
            col_index_right=end_col,
            col_index_left=end_col - config.primer_size_min,
            row_index=row_index,
            seq_str=start_seq,
            config=config,
        )
        if min_freq == 0 and set(results) & ERROR_SET:
            return (end_col, {parse_error(set(results)): -1})

        # Add the results to the Counter
        total_col_seqs.update(
            {seq: 1 / len(results) for seq in parse_error_list(results)}
        )

    return (end_col, dict(total_col_seqs))


def f_digest_index(
    align_array: np.ndarray, config: Config, end_col: int, min_freq: float
) -> FKmer | tuple[int, DIGESTION_ERROR]:
    """
    This will try and create a FKmer ended at the given index
    :align_array: The alignment array
    :config: The configuration object
    :end_col: The column index to end the FKmer
    :min_freq: The minimum frequency threshold

    :return: A FKmer object or a tuple of (end_col, error)
    """

    # Count how many times each sequence / error occurs
    _end_col, seq_counts = f_digest_to_count(align_array, config, end_col, min_freq)
    tmp_parsed_seqs = process_seqs(seq_counts, min_freq, ignore_n=config.ignore_n)
    if isinstance(tmp_parsed_seqs, DIGESTION_ERROR):
        return (end_col, tmp_parsed_seqs)
    elif isinstance(tmp_parsed_seqs, dict):
        parsed_seqs = tmp_parsed_seqs
    else:
        raise ValueError("Unknown error occurred")

    # Thermo check the kmers
    thermo_result = thermo_check_kmers({*parsed_seqs.keys()}, config)
    match thermo_result:
        case THERMORESULT.PASS:
            pass
        case _:
            return (end_col, parse_thermo_error(thermo_result))

    if do_pools_interact_py(
        [*parsed_seqs.keys()], [*parsed_seqs.keys()], config.dimer_score
    ):
        return (end_col, DIGESTION_ERROR.DIMER_FAIL)

    if not parsed_seqs:
        return (end_col, DIGESTION_ERROR.NO_SEQUENCES)

    return FKmer(end_col, list(parsed_seqs.keys()))


def hamming_dist(s1, s2) -> int:
    """
    Return the number of substitutions, starting from the 3p end
    """
    return sum((x != y for x, y in zip(s1[::-1], s2[::-1])))


def f_digest(
    msa_array: np.ndarray, config: Config, findexes: list[int], logger
) -> list[FKmer]:
    fkmers = []
    for findex in findexes:
        fkmer = f_digest_index(msa_array, config, findex, config.min_base_freq)

        # Append valid FKmers
        if isinstance(fkmer, FKmer) and fkmer.seqs:
            fkmers.append(fkmer)

        # Log the Digestion
        if logger is not None:
            if isinstance(fkmer, tuple):
                logger.debug(f"FKmer: [red]{fkmer[0]}[/red]\t{fkmer[1].value}")
            else:
                logger.debug(f"FKmer: [green]{fkmer.end}[/green]: AllPass")

    return fkmers


def r_digest(
    msa_array: np.ndarray, config: Config, rindexes: list[int], logger
) -> list[RKmer]:
    rkmers = []
    for rindex in rindexes:
        rkmer = r_digest_index(msa_array, config, rindex, config.min_base_freq)

        # Append valid RKmers
        if isinstance(rkmer, RKmer) and rkmer.seqs:
            rkmers.append(rkmer)

        # Log the Digestion
        if logger is not None:
            if isinstance(rkmer, tuple):
                logger.debug(f"RKmer: [red]{rkmer[0]}[/red]\t{rkmer[1].value}")
            else:
                logger.debug(f"RKmer: [green]{rkmer.start}[/green]: AllPass")
    return rkmers


def digest(
    msa_array: np.ndarray,
    config: Config,
    progress_manager: ProgressManager,
    indexes: tuple[list[int], list[int]] | None = None,
    logger=None,
    chrom: str = "",
) -> tuple[list[FKmer], list[RKmer]]:
    """
    Digest the given MSA array and return the FKmers and RKmers.

    :param msa_array: The input MSA array.
    :param cfg: A dictionary containing configuration parameters.
    :param indexes: A tuple of MSA indexes for (FKmers, RKmers), or None to use all indexes.
    :param logger: None or the logger object.
    :return: A tuple containing lists of sorted FKmers and RKmers.
    """
    # Guard for invalid indexes
    if indexes is not None:
        if min(indexes[0]) < 0 or max(indexes[0]) >= msa_array.shape[1]:
            raise IndexError("FIndexes are out of range")
        if min(indexes[1]) < 0 or max(indexes[1]) >= msa_array.shape[1]:
            raise IndexError("RIndexes are out of range")

    # Get the indexes to digest
    findexes = (
        indexes[0]
        if indexes is not None
        else range(config.primer_size_min, msa_array.shape[1])
    )
    rindexes = (
        indexes[1]
        if indexes is not None
        else range(msa_array.shape[1] - config.primer_size_min)
    )

    # Digest the findexes
    fkmers = []
    pt = progress_manager.create_sub_progress(
        iter=findexes, process="Creating forward primers", chrom=chrom
    )
    for findex in pt:
        fkmer = f_digest_index(msa_array, config, findex, config.min_base_freq)

        if logger is not None:
            if isinstance(fkmer, tuple):
                logger.debug(f"{chrom}:FKmer: {fkmer[0]}\t{fkmer[1].value}")
            else:
                logger.debug(f"{chrom}:FKmer: {fkmer.end}\tPass")

        # Append valid FKmers
        if isinstance(fkmer, FKmer) and fkmer.seqs:
            fkmers.append(fkmer)

        # Update the count
        pt.manual_update(count=len(fkmers))

    # Digest the rindexes
    rkmers = []
    pt = progress_manager.create_sub_progress(
        iter=rindexes, process="Creating reverse primers", chrom=chrom
    )
    for rindex in pt:
        rkmer = r_digest_index(msa_array, config, rindex, config.min_base_freq)

        if logger is not None:
            if isinstance(rkmer, tuple):
                logger.debug(f"{chrom}:RKmer: {rkmer[0]}\t{rkmer[1].value}")
            else:
                logger.debug(f"{chrom}:RKmer: {rkmer.start}\tPass")

        # Append valid RKmers
        if isinstance(rkmer, RKmer) and rkmer.seqs:
            rkmers.append(rkmer)

        # Update the count
        pt.manual_update(count=len(rkmers))

    return (fkmers, rkmers)



import pathlib
from enum import Enum
from typing import Any

from primalscheme3 import __version__


# Written by Andy Smith, modified by: Chris Kent
class MappingType(Enum):
    """
    Enum for the mapping type
    """

    FIRST = "first"
    CONSENSUS = "consensus"


class Config:
    """
    PrimalScheme3 configuration.
    Class properties are defaults, can be overridden
    on instantiation (and will shadow class defaults)
    """

    # Run Settings
    output: pathlib.Path = pathlib.Path("./output")
    force: bool = False
    high_gc: bool = False
    input_bedfile: pathlib.Path | None = None
    version: str = __version__
    # Scheme Settings
    n_pools: int = 2
    min_overlap: int = 10
    mapping: MappingType = MappingType.FIRST
    circular: bool = False
    backtrack: bool = False
    min_base_freq: float = 0.0
    ignore_n: bool = False
    # Amplicon Settings
    amplicon_size: int = 400
    amplicon_size_min: int = 0
    amplicon_size_max: int = 0
    # Primer Settings
    _primer_size_default_min: int = 19
    _primer_size_default_max: int = 34
    _primer_size_hgc_min: int = 17
    _primer_size_hgc_max: int = 30
    _primer_gc_default_min: int = 30
    _primer_gc_default_max: int = 55
    _primer_gc_hgc_min: int = 40
    _primer_gc_hgc_max: int = 65
    primer_tm_min: float = 59.5
    primer_tm_max: float = 62.5
    primer_hairpin_th_max: float = 47.0
    primer_homopolymer_max: int = 5
    primer_max_walk: int = 80
    # MatchDB Settings
    use_matchdb: bool = True
    editdist_max: int = 1
    mismatch_fuzzy: bool = True
    mismatch_kmersize: int  # Same as primer_size_min
    mismatch_product_size: int = 0
    # Thermodynamic Parameters
    mv_conc: float = 100.0
    dv_conc: float = 2.0
    dntp_conc: float = 0.8
    dna_conc: float = 15.0
    dimer_score: float = -26.0

    def __init__(self, **kwargs: Any) -> None:
        self.assign_kwargs(**kwargs)
        # Set amplicon size
        if self.amplicon_size_min == 0:
            self.amplicon_size_min = int(self.amplicon_size * 0.9)
        if self.amplicon_size_max == 0:
            self.amplicon_size_max = int(self.amplicon_size * 1.1)
        if self.high_gc:
            self.primer_size_min = self._primer_size_hgc_min
            self.primer_size_max = self._primer_size_hgc_max
            self.primer_gc_min = self._primer_gc_hgc_min
            self.primer_gc_max = self._primer_gc_hgc_max
        else:
            self.primer_size_min = self._primer_size_default_min
            self.primer_size_max = self._primer_size_default_max
            self.primer_gc_min = self._primer_gc_default_min
            self.primer_gc_max = self._primer_gc_default_max
        # Set MisMatch Kmer Size
        self.mismatch_kmersize = self.primer_size_min

    def items(self) -> dict[str, Any]:
        """
        Return a dict (key, val) for non-private, non-callable members
        """
        items = {}
        for key in [x for x in dir(self) if not x.startswith("_")]:
            if not callable(getattr(self, key)):  # prevent functions
                value = getattr(self, key)
                # Convert Enum and Path objects to their values
                if isinstance(value, Enum):
                    items[key] = value.value
                if isinstance(value, pathlib.Path):
                    items[key] = str(value)
                else:
                    items[key] = value

        return items

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dict (key, val) for non-private, non-callable members
        """
        dict = {}
        for key, val in self.items().items():
            if isinstance(val, Enum):
                dict[key] = val.value
            elif isinstance(val, pathlib.Path):
                dict[key] = str(val)
            else:
                dict[key] = val
        return dict

    def __str__(self) -> str:
        return "\n".join(f"{key}: {val}" for key, val in self.items())

    def assign_kwargs(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            # Check if key is valid
            if value is None:
                continue

            if key == "input_bedfile":
                setattr(self, key, pathlib.Path(value))
                continue

            if hasattr(self, key):
                # Convert to expected type
                if isinstance(getattr(self, key), MappingType):
                    setattr(self, key, MappingType(value))
                elif isinstance(getattr(self, key), pathlib.Path):
                    setattr(self, key, pathlib.Path(value))
                elif isinstance(
                    getattr(self, key), bool
                ):  # Need to check bool before int
                    parsed_bool = str(value).lower()
                    setattr(self, key, parsed_bool == "true")
                elif isinstance(getattr(self, key), int):
                    setattr(self, key, int(value))
                elif isinstance(getattr(self, key), float):
                    setattr(self, key, float(value))
                elif isinstance(getattr(self, key), str):
                    setattr(self, key, str(value))
                else:
                    print(f"Could not parse {key} with value {value} ({type(value)}")


# All bases allowed in the input MSA
IUPAC_ALL_ALLOWED_DNA = {
    "A",
    "G",
    "K",
    "Y",
    "B",
    "S",
    "N",
    "H",
    "C",
    "W",
    "D",
    "R",
    "M",
    "T",
    "V",
    "-",
}

SIMPLE_BASES = {"A", "C", "G", "T"}

AMBIGUOUS_DNA = {
    "M": "AC",
    "R": "AG",
    "W": "AT",
    "S": "CG",
    "Y": "CT",
    "K": "GT",
    "V": "ACG",
    "H": "ACT",
    "D": "AGT",
    "B": "CGT",
}
ALL_DNA: dict[str, str] = {
    "A": "A",
    "C": "C",
    "G": "G",
    "T": "T",
    "M": "AC",
    "R": "AG",
    "W": "AT",
    "S": "CG",
    "Y": "CT",
    "K": "GT",
    "V": "ACG",
    "H": "ACT",
    "D": "AGT",
    "B": "CGT",
}
ALL_BASES: set[str] = {
    "A",
    "C",
    "G",
    "T",
    "M",
    "R",
    "W",
    "S",
    "Y",
    "K",
    "V",
    "H",
    "D",
    "B",
}
AMB_BASES = {"Y", "W", "R", "B", "H", "V", "D", "K", "M", "S"}
AMBIGUOUS_DNA_COMPLEMENT = {
    "A": "T",
    "C": "G",
    "G": "C",
    "T": "A",
    "M": "K",
    "R": "Y",
    "W": "W",
    "S": "S",
    "Y": "R",
    "K": "M",
    "V": "B",
    "H": "D",
    "D": "H",
    "B": "V",
    "X": "X",
    "N": "N",
    "-": "-",
}



import pathlib
from enum import Enum
from itertools import islice

import numpy as np

# Iterations checker
from primalscheme3.core.bedfiles import BedPrimerPair

# Core Module imports
from primalscheme3.core.classes import PrimerPair
from primalscheme3.core.config import Config
from primalscheme3.core.errors import CustomErrors
from primalscheme3.core.mismatches import MatchDB
from primalscheme3.core.msa import MSA
from primalscheme3.core.multiplex import Multiplex, PrimerPairCheck
from primalscheme3.core.progress_tracker import ProgressManager
from primalscheme3.core.seq_functions import (
    entropy_score_array,
)


class PanelRunModes(Enum):
    ENTROPY = "entropy"
    REGION_ONLY = "region-only"
    EQUAL = "equal"


class PanelReturn(Enum):
    """
    Enum for the return values of the Panel class.
    """

    ADDED_PRIMERPAIR = 0  # A primerpair was added
    NO_PRIMERPAIRS = 1  # No more primerpairs to add (finished)
    NO_PRIMERPAIRS_IN_MSA = 3  # No more primerpairs in the current MSA (moving to next)


def does_overlap(
    new_pp: tuple[int, int, int], current_pps: list[tuple[int, int, int]]
) -> bool:
    """Checks if a new primerpair overlaps with any of the current primerpair.

    Args:
        new_pp: A tuple representing the new primerpair to check for overlap. The tuple
            contains three integers: the start min(pp.fprimer.starts) , end max(pp.rprimer.ends), and the MSA (multiple sequence
            alignment) index.
        current_pps: A list of tuples representing the current primal panels to check for
            overlap. Each tuple contains three integers: the start index, end index, and the
            MSA index.

    Returns:
        A boolean value indicating whether the new primal panel overlaps with any of the current
        primal panels.
    """
    # Check if from same msa
    for current_pp in current_pps:
        # if not from same msa
        if current_pp[2] != new_pp[2]:
            continue

        # If they overlap
        if range(max(new_pp[0], current_pp[0]), min(new_pp[1], current_pp[1]) + 1):
            return True

    return False


class Region:
    chromname: str
    start: int
    stop: int
    name: str
    score: int
    group: str | None

    def __init__(
        self,
        chromname: str,
        start: int,
        stop: int,
        name: str,
        score: int,
        group: str | None = None,
    ) -> None:
        self.chromname = str(chromname).strip()
        self.start = int(start)
        self.stop = int(stop)
        self.name = str(name).strip()
        self.score = int(score)
        if self.start >= self.stop:
            raise ValueError(f"{self.name}: Circular regions are not supported.")

        # Parse Group
        if group is not None and group != "":
            group = str(group).strip()
        else:
            group = None

        self.group = group

    def positions(self):
        return range(self.start, self.stop)

    def __hash__(self) -> int:
        return hash(
            f"{self.chromname}:{self.start}:{self.stop}:{self.name}:{self.score}"
        )

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Region):
            return False
        return hash(self) == hash(__value)

    def to_bed(self) -> str:
        group_str = self.group if self.group else ""
        return f"{self.chromname}\t{self.start}\t{self.stop}\t{self.name}\t{self.score}\t{group_str}\n"


class RegionParser:
    @staticmethod
    def from_list(bed_list: list[str]) -> Region:
        """
        Returns a single region object from a list representing a line of a bedfile, split by tabs
        """
        # Handle extra columns
        if len(bed_list) >= 5:
            chromname, start, stop, name, score = bed_list[:5]
        else:
            raise CustomErrors(
                f"Invalid region: {' '.join(bed_list)}. Requires 5 or more columns."
            )

        try:
            start = int(start)
            stop = int(stop)
            score = int(score)
        except ValueError as e:
            raise CustomErrors(
                f"Invalid region: {' '.join(bed_list)}. Start, stop, score must be integers."
            ) from e

        # Handle group column
        try:
            group = bed_list[5]
            if group == "" or not group:  # Allow empty group
                group = None
        except IndexError:
            group = None

        return Region(chromname, start, stop, name, score, group)

    @staticmethod
    def from_str(bed_str: str) -> Region:
        """
        Returns a single region object from a string representing a line of a bedfile
        """
        return RegionParser.from_list(bed_str.strip().split("\t"))


class PanelMSA(MSA):
    # Score arrays
    _score_array: np.ndarray | None
    _midx_entropy_array: np.ndarray | None
    regions: list[Region] | None
    region_group_count: dict[str, int] | None
    primerpairpointer: int

    def __init__(
        self,
        name: str,
        path: pathlib.Path,
        msa_index: int,
        mapping: str,
        progress_manager: ProgressManager,
        logger=None,
    ) -> None:
        # Call the MSA init
        super().__init__(
            name=name,
            path=path,
            msa_index=msa_index,
            mapping=mapping,
            logger=logger,
            progress_manager=progress_manager,
        )

        # Create the primerpairpointer
        self.primerpairpointer = 0
        self.regions = None

    def create_score_array(
        self,
        regions: list[Region] | None,
        mode: PanelRunModes = PanelRunModes.REGION_ONLY,
    ) -> None:
        match mode:
            case PanelRunModes.REGION_ONLY:
                if regions is None:
                    raise ValueError("Regions must be provided to create score array.")
                self._score_array = np.zeros(len(self._mapping_array), dtype=int)
                for region in regions:
                    self._score_array[region.start : region.stop] += region.score
            case PanelRunModes.ENTROPY:
                # Turn the msa indexed entropy array into a ref indexed array
                msa_entropy_array = entropy_score_array(self.array)
                self._score_array = np.array(
                    [
                        x + 0.01  # Add a small value to avoid 0
                        for msai, x in enumerate(msa_entropy_array)
                        if self._mapping_array[msai] is not None
                    ]
                )
            case PanelRunModes.EQUAL:
                self._score_array = np.ones(len(self._mapping_array), dtype=int)

    def create_entropy_array(self):
        self._midx_entropy_array = np.array(entropy_score_array(self.array))

    def add_regions(self, regions: list[Region] | None) -> None:
        self.regions = regions
        if regions is not None:
            # Create an empty dict
            region_groups = {x.group for x in regions if x.group}
            self.region_group_count = {x: 0 for x in region_groups}

    def remove_kmers_that_clash_with_regions(self):
        """
        Removes f/rkmers who clash with the regions
        """
        if self.regions is None:
            return

        # Remove primer that overlap with regions
        regions = [(x.start, x.stop, self.msa_index) for x in self.regions]
        self.fkmers = [
            fkmer
            for fkmer in self.fkmers
            if not does_overlap(
                (min(fkmer.starts()), fkmer.end, self.msa_index),
                regions,
            )
        ]
        self.rkmers = [
            rkmer
            for rkmer in self.rkmers
            if not does_overlap(
                (rkmer.start, max(rkmer.ends()), self.msa_index),
                regions,
            )
        ]

    def iter_unchecked_primerpairs(self):
        """
        Returns all primerpairs that have not been checked yet
        """
        return islice(self.primerpairs, self.primerpairpointer, None)

    def get_pp_entropy(self, pp: PrimerPair) -> float:
        """
        Returns sum of entropy in the primertrimmed amplicon
        """
        assert self._midx_entropy_array is not None
        ppptstart, ppptend = pp.primertrimmed_region()

        return np.sum(
            self._midx_entropy_array[
                self._ref_to_msa[ppptstart] : self._ref_to_msa[ppptend]
            ]
        )

    def get_pp_score(self, pp: PrimerPair) -> int:
        """
        Returns number of SNPs in the primertrimmed amplicon
        """
        assert self._score_array is not None
        ppptstart, ppptend = pp.primertrimmed_region()

        # circular
        if pp.fprimer.end > pp.rprimer.start:
            return np.sum(self._score_array[ppptstart:]) + np.sum(
                self._score_array[:ppptend]
            )
        else:
            return np.sum(self._score_array[ppptstart:ppptend])

    def update_score_array(self, addedpp: PrimerPair, newscore: int = 0) -> None:
        """
        Updates the score array with the added primerpair
        """
        assert self._score_array is not None
        ppptstart, ppptend = addedpp.primertrimmed_region()
        # circular
        if addedpp.fprimer.end > addedpp.rprimer.start:
            self._score_array[ppptstart:] = newscore
            self._score_array[:ppptend] = newscore
        else:
            self._score_array[ppptstart:ppptend] = newscore


class Panel(Multiplex):
    # Base class
    _pools: list[list[PrimerPair | BedPrimerPair]]
    _current_pool: int
    _last_pp_added: list[PrimerPair]  # Stack to keep track of the last primer added
    _matchDB: MatchDB
    config: Config
    _msa_dict: dict[int, PanelMSA]

    # New attributes
    _current_msa_index: int

    # Keep adding
    _is_msa_index_finished: dict[int, bool]

    def __init__(
        self,
        msa_dict: dict[int, PanelMSA],
        config: Config,
        matchdb: MatchDB,
        logger=None,
    ) -> None:
        super().__init__(config, matchdb, msa_dict)  # type: ignore

        self._current_msa_index = 0
        self._failed_primerpairs = [set() for _ in range(self.n_pools)]

        # Keep adding
        self._is_msa_index_finished = {msa_index: False for msa_index in msa_dict}

        self.logger = logger

    def _next_msa(self) -> int:
        """
        Updates the current msa index to the next msa. Returns the new msa index.
        :return: The new msa index.
        """
        self._current_msa_index = (self._current_msa_index + 1) % len(self._msa_dict)
        return self._current_msa_index

    def _add_primerpair(
        self, primerpair: PrimerPair, pool: int, msa_index: int
    ) -> None:
        # Add the primerpair to the pool
        super().add_primer_pair_to_pool(primerpair, pool, msa_index)

        # Update the msa Score array
        if msa_index in self._msa_dict:
            self._msa_dict[msa_index].update_score_array(primerpair)

        # Update the current MSA
        self._next_msa()

    def add_next_primerpair(
        self, max_amplicons_group: None | int = None
    ) -> PanelReturn:
        """
        Try and add the next primerpair
        """
        # Get current MSA
        current_msa = self._msa_dict[self._current_msa_index]
        current_pool = self._current_pool

        # Check if the current MSA is finished
        if self._is_msa_index_finished[self._current_msa_index]:
            self._next_msa()
            return PanelReturn.NO_PRIMERPAIRS_IN_MSA

        # Pools to try
        pos_pools_indexes = [
            (current_pool + i) % self.n_pools for i in range(self.n_pools)
        ]

        # All seqs in each pool
        seqs_in_each_pool = {
            pos_pool: [
                seq
                for seq in (pp.all_seqs() for pp in self._pools[pos_pool])
                for seq in seq
            ]
            for pos_pool in pos_pools_indexes
        }

        # Remove primerpairs with no score
        # current_msa.primerpairs = [
        #     pp for pp in current_msa.primerpairs if current_msa.get_pp_score(pp) > 0
        # ]
        # For the primerpairs in the current MSA Sort all primerpairs by score
        current_msa.primerpairs.sort(
            key=lambda x: (
                self.calc_pp_score(current_msa, x),
                -len(x.all_seqs()),
            ),
            reverse=True,
        )
        for pospp in current_msa.primerpairs:
            # Check primer has score
            if self.calc_pp_score(current_msa, pospp, False) <= 0:
                continue
            for pospool in pos_pools_indexes:
                # Check if the primerpair can be added
                match self.check_primerpair_can_be_added(
                    pospp, pospool, seqs_in_each_pool[pospool]
                ):
                    case PrimerPairCheck.OK:
                        # _add_primerpair updates the score array for the added primerpair
                        self._add_primerpair(pospp, pospool, self._current_msa_index)
                        # Check for other regions in the same group
                        # Find regions with overlap
                        if current_msa.regions:
                            # Find other regions in the same group
                            other_regions = self.get_regions_with_group_overlap()
                            # Update the MSA region_group_count
                            for region in other_regions:
                                assert current_msa.region_group_count is not None
                                assert current_msa._score_array is not None
                                assert region.group is not None

                                if max_amplicons_group is not None and (
                                    current_msa.region_group_count[region.group]
                                    >= max_amplicons_group
                                ):  # type: ignore
                                    current_msa._score_array[
                                        region.start : region.stop
                                    ] = 0

                                current_msa.region_group_count[region.group] += 1  # type: ignore

                        return PanelReturn.ADDED_PRIMERPAIR
                    case _:
                        continue
        # No more primerpairs in the current MSA
        self._is_msa_index_finished[self._current_msa_index] = True
        return PanelReturn.NO_PRIMERPAIRS_IN_MSA

    def calc_pp_score(self, msa: PanelMSA, pp: PrimerPair, bonus_ol=True) -> int:
        """
        Returns number of SNPs in the primertrimmed amplicon
        """
        # Get the score
        score = msa.get_pp_score(pp)

        # Apply a bonus for good gc content
        gc_diff = pp.get_score()  # 0-0.5
        score = score * (1 - gc_diff)

        # Apply a bonus if the primerpair spans a coverage break
        if bonus_ol:
            coverage_slice = self._coverage[msa.msa_index][
                pp.fprimer.end : pp.rprimer.start
            ]
            if True in coverage_slice and False in coverage_slice:
                score = score * 10

        return int(score)

    def get_regions_with_group_overlap(self) -> list[Region]:
        """
        Finds all regions groups which are overlapping with the last primerpair added.
        Returns all regions with the same groups.
        """
        last_primerpair = self._last_pp_added[-1]
        current_msa = self._msa_dict[self._current_msa_index]
        assert current_msa.regions is not None

        ol_regions_groups = {
            x.group
            for x in current_msa.regions
            if x.group
            and does_overlap(
                (
                    last_primerpair.fprimer.region()[0],
                    last_primerpair.rprimer.region()[1],
                    self._current_msa_index,
                ),
                [(x.start, x.stop, self._current_msa_index)],
            )
        }
        other_regions = [x for x in current_msa.regions if x.group in ol_regions_groups]
        return other_regions



import hashlib
import json
import pathlib
import shutil
from enum import Enum

from click import UsageError
from primaldimer_py import do_pools_interact_py  # type: ignore

# Core imports
from primalscheme3.core.bedfiles import read_in_bedprimerpairs
from primalscheme3.core.config import Config
from primalscheme3.core.digestion import (
    DIGESTION_ERROR,
    f_digest_to_count,
    r_digest_to_count,
)
from primalscheme3.core.logger import setup_rich_logger
from primalscheme3.core.msa import MSA
from primalscheme3.core.progress_tracker import ProgressManager
from primalscheme3.core.seq_functions import reverse_complement
from primalscheme3.core.thermo import THERMORESULT, thermo_check


class NewPrimerStatus(Enum):
    VALID = "valid"
    PRESENT = "present"
    FAILED = "failed"


class SeqStatus:
    seq: str | None
    count: int
    thermo_status: THERMORESULT | DIGESTION_ERROR

    def __init__(
        self, seq: str | None, count: int, thermo_status: THERMORESULT | DIGESTION_ERROR
    ):
        self.seq = seq
        self.count = count
        self.thermo_status = thermo_status

    def __str__(self) -> str:
        return f"{self.seq}\t{self.count}\t{self.thermo_status}"


def detect_early_return(seq_counts: dict[str | DIGESTION_ERROR, int]) -> bool:
    """
    Checks for an early return condition, will return True condition is met
    """
    # Check for early return conditions
    for error, count in seq_counts.items():
        if count == -1 and type(error) == DIGESTION_ERROR:
            return True
    return False


def report_check(
    seqstatus: SeqStatus,
    current_primer_seqs: set[str],
    seqs_in_pools: list[list[str]],
    pool: int,
    dimerscore: float,
    logger,
    config: Config,
) -> bool:
    """
    Will carry out the checks and report the results via the logger. Will return False if the seq should not be added
    """

    report_seq = seqstatus.seq if seqstatus.seq is not None else "DIGESTION_ERROR"
    report_seq = report_seq.rjust(config.primer_size_max + 5, " ")

    # Check it passed thermo
    if seqstatus.thermo_status != THERMORESULT.PASS or seqstatus.seq is None:
        logger.warning(
            f"{report_seq}\t{seqstatus.count}\t[red]{NewPrimerStatus.FAILED.value}[/red]: {seqstatus.thermo_status}",
        )
        return False

    # Check it is a new seq
    if seqstatus.seq in current_primer_seqs:
        logger.info(
            f"{report_seq}\t{seqstatus.count}\t[blue]{NewPrimerStatus.PRESENT.value}[/blue]: In scheme",
        )
        return False

    # Check for minor allele
    if seqstatus.count < 0:
        logger.warning(
            f"{report_seq}\t{seqstatus.count}\t[red]{NewPrimerStatus.FAILED.value}[/red]: Minor allele",
        )
        return False

    # Check for dimer with pool
    if do_pools_interact_py([seqstatus.seq], seqs_in_pools[pool], dimerscore):
        logger.warning(
            f"{report_seq}\t{seqstatus.count}\t[red]{NewPrimerStatus.FAILED.value}[/red]: Interaction with pool",
        )
        return False

    # Log the seq
    logger.info(
        f"{report_seq}\t{seqstatus.count}\t[green]{NewPrimerStatus.VALID.value}[/green]: Can be added",
    )

    return True


def repair(
    config_path: pathlib.Path,
    msa_path: pathlib.Path,
    bedfile_path: pathlib.Path,
    output_dir: pathlib.Path,
    force: bool,
    pm: ProgressManager | None,
):
    OUTPUT_DIR = pathlib.Path(output_dir).absolute()  # Keep absolute path

    # Read in the config file
    with open(config_path) as f:
        base_cfg = json.load(f)

    msa_data = base_cfg["msa_data"]

    # Parse params from the config
    config = Config(**base_cfg)
    base_cfg = config.to_dict()
    base_cfg["msa_data"] = msa_data

    config.min_base_freq = 0.01

    # See if the output dir already exists
    if OUTPUT_DIR.is_dir() and not force:
        raise UsageError(f"{OUTPUT_DIR} already exists, please use --force to override")

    # Create the output dir and a work subdir
    pathlib.Path.mkdir(OUTPUT_DIR, exist_ok=True)
    pathlib.Path.mkdir(OUTPUT_DIR / "work", exist_ok=True)

    ## Set up the logger
    logger = setup_rich_logger(str(OUTPUT_DIR / "work" / "file.log"))

    ## Set up the progress manager
    if pm is None:
        pm = ProgressManager()

    # Read in the MSA file
    msa_obj = MSA(
        name=msa_path.stem,
        path=msa_path,
        msa_index=0,
        mapping=base_cfg["mapping"],
        logger=logger,
        progress_manager=pm,
    )
    logger.info(
        f"Read in MSA: [blue]{msa_path.name}[/blue] ({msa_obj._chrom_name})\t"
        f"seqs:[green]{msa_obj.array.shape[0]}[/green]\t"
        f"cols:[green]{msa_obj.array.shape[1]}[/green]"
    )
    # Check for a '/' in the chromname
    if "/" in msa_obj._chrom_name:
        new_chromname = msa_obj._chrom_name.split("/")[0]
        logger.warning(
            f"Having a '/' in the chromname {msa_obj._chrom_name} "
            f"will cause issues with figure generation bedfile output. "
            f"Parsing chromname [yellow]{msa_obj._chrom_name}[/yellow] -> [green]{new_chromname}[/green]"
        )
        msa_obj._chrom_name = new_chromname

    # Update the base_cfg with the new msa
    # Create MSA checksum
    with open(msa_path, "rb") as f:
        msa_checksum = hashlib.file_digest(f, "md5").hexdigest()

    current_msa_index = max([int(x) for x in base_cfg["msa_data"].keys()])
    base_cfg["msa_data"][str(current_msa_index + 1)] = {
        "msa_name": msa_obj.name,
        "msa_path": str("work/" + msa_path.name),
        "msa_chromname": msa_obj._chrom_name,
        "msa_uuid": msa_obj._uuid,
        "msa_checksum": msa_checksum,
    }
    # Copy the MSA file to the work dir
    local_msa_path = OUTPUT_DIR / "work" / msa_path.name
    shutil.copy(msa_path, local_msa_path)

    # Read in the bedfile
    all_primerpairs, _header = read_in_bedprimerpairs(bedfile_path)

    # Get the primerpairs for this new MSA
    primerpairs_in_msa = [
        pp for pp in all_primerpairs if pp.chrom_name == msa_obj._chrom_name
    ]

    if len(primerpairs_in_msa) == 0:
        logger.critical(
            f"No primerpairs found for {msa_obj._chrom_name} in {bedfile_path}",
        )
        raise UsageError(
            f"No primerpairs found for {msa_obj._chrom_name} in {bedfile_path}"
        )

    # Get all the seqs in each pool
    seqs_in_pools = [[] for _ in range(config.n_pools)]
    for pp in primerpairs_in_msa:
        seqs_in_pools[pp.pool].extend([*pp.fprimer.seqs, *pp.rprimer.seqs])

    # Find the indexes in the MSA that the primerbed refer to
    assert msa_obj._mapping_array is not None
    mapping_list = list(msa_obj._mapping_array)

    # For primerpair in the bedfile, check if new seqs need to be added by digestion the MSA
    for pp in primerpairs_in_msa:
        logger.info(
            f"Checking {pp.amplicon_prefix}_{pp.amplicon_number}_LEFT",
        )
        msa_fkmer_end = msa_obj._ref_to_msa.get(pp.fprimer.end)

        if msa_fkmer_end is None:
            continue

        _end_col, fseq_counts = f_digest_to_count(
            msa_obj.array, config, msa_fkmer_end, config.min_base_freq
        )

        # Check for early return conditions
        if detect_early_return(fseq_counts):
            logger.warning(
                f"Early return for {pp.amplicon_prefix}_{pp.amplicon_number}_LEFT. Skipping",
            )
            continue

        # Thermo check each sequence
        seqstatuss: list[SeqStatus] = []
        for seq, count in fseq_counts.items():
            if isinstance(seq, DIGESTION_ERROR):
                thermo_status = seq
                seq = None
            else:
                thermo_status = thermo_check(seq, config=config)
            seqstatuss.append(SeqStatus(seq, count, thermo_status))
        seqstatuss = sorted(seqstatuss, key=lambda x: x.count, reverse=True)

        # Decide if the new seqs should be added
        for seqstatus in seqstatuss:
            if not report_check(
                seqstatus=seqstatus,
                current_primer_seqs=pp.fprimer.seqs,
                seqs_in_pools=seqs_in_pools,
                pool=pp.pool,
                dimerscore=config.dimer_score,
                logger=logger,
                config=config,
            ):
                continue

            # Add the new seq
            seqs_in_pools[pp.pool].append(seqstatus.seq)

        # Handle the right primer
        logger.info(
            f"Checking {pp.amplicon_prefix}_{pp.amplicon_number}_RIGHT",
        )
        msa_rkmer_start = mapping_list.index(pp.rprimer.start)
        _start_col, rseq_counts = r_digest_to_count(
            msa_obj.array, config, msa_rkmer_start, config.min_base_freq
        )
        # Check for early return conditions
        if detect_early_return(rseq_counts):
            logger.warning(
                "Early return for {pp.amplicon_prefix}_{pp.amplicon_number}_RIGHT",
            )
            continue

        # Valid seqs
        valid_rseqs = {
            reverse_complement(seq): count
            for seq, count in rseq_counts.items()
            if isinstance(seq, str)
        }

        # Thermo check each sequence
        rseqstatuss: list[SeqStatus] = []
        for seq, count in valid_rseqs.items():
            if isinstance(seq, DIGESTION_ERROR):
                thermo_status = seq
                seq = None
            else:
                thermo_status = thermo_check(seq, config=config)
            rseqstatuss.append(SeqStatus(seq, count, thermo_status))
        rseqstatuss = sorted(rseqstatuss, key=lambda x: x.count, reverse=True)

        # Decide if the new seqs should be added
        for rseqstatus in rseqstatuss:
            if not report_check(
                seqstatus=rseqstatus,
                current_primer_seqs=pp.rprimer.seqs,
                seqs_in_pools=seqs_in_pools,
                pool=pp.pool,
                dimerscore=config.dimer_score,
                logger=logger,
                config=config,
            ):
                continue

            # Add the new seq
            seqs_in_pools[pp.pool].append(rseqstatus.seq)

    # Write out the new bedfile
    with open(OUTPUT_DIR / "primer.bed", "w") as f:
        for pp in primerpairs_in_msa:
            pp.amplicon_prefix = msa_obj._uuid
            f.write(pp.to_bed() + "\n")

    # Amplicon and primertrimmed files should not have changed. Can be copied from the input dir
    # Not sure how to handle the amplicon names, as the primerstem has changed?
    ## Keep original names for now

    # Write the config dict to file
    with open(OUTPUT_DIR / "config.json", "w") as outfile:
        outfile.write(json.dumps(base_cfg, sort_keys=True))



from enum import Enum

from primaldimer_py import (
    do_pools_interact_py,  # type: ignore
    which_kmers_pools_interact,  # type: ignore
)

from primalscheme3.core.bedfiles import BedPrimerPair
from primalscheme3.core.classes import PrimerPair
from primalscheme3.core.config import Config
from primalscheme3.core.get_window import get_pp_window
from primalscheme3.core.mismatches import MatchDB, detect_new_products
from primalscheme3.core.msa import MSA
from primalscheme3.core.multiplex import Multiplex, PrimerPairCheck

# Scheme imports
from primalscheme3.scheme.primer_pair_score import (
    bt_ol_pp_score,
    ol_pp_score,
    walk_pp_score,
)


class SchemeReturn(Enum):
    # Added return values
    ADDED_OL_PRIMERPAIR = 1
    ADDED_WALK_PRIMERPAIR = 2
    ADDED_FIRST_PRIMERPAIR = 3
    # Failed return values
    NO_OL_PRIMERPAIR = 4
    NO_WALK_PRIMERPAIR = 5
    NO_FIRST_PRIMERPAIR = 6
    # Misc return values
    ADDED_BACKTRACKED = 7
    NO_BACKTRACK = 8

    ADDED_CIRCULAR = 9
    NO_CIRCULAR = 10


class Scheme(Multiplex):
    _pools: list[list[PrimerPair | BedPrimerPair]]
    _current_pool: int
    _last_pp_added: list[PrimerPair]  # Stack to keep track of the last primer added
    _matchDB: MatchDB
    config: Config

    def __init__(self, config, matchDB: MatchDB, msa_dict: dict[int, MSA]):
        super().__init__(config, matchDB, msa_dict)

    @property
    def npools(self) -> int:
        return self.n_pools

    def add_first_primer_pair(
        self, primerpairs: list[PrimerPair], msa_index
    ) -> SchemeReturn:
        "Adds primerpair to the current pool, and updates the current pool"
        # If there are no primerpairs, return false
        if not primerpairs:
            return SchemeReturn.NO_FIRST_PRIMERPAIR

        # Try and add the first primerpair to an empty pool
        for pool_index in range(self.n_pools):
            if not self._pools[pool_index]:
                self.add_primer_pair_to_pool(primerpairs[0], pool_index, msa_index)
                return SchemeReturn.ADDED_FIRST_PRIMERPAIR

        # Create a hashmap of what seqs are in each pool for quick look up
        pool_seqs_map: dict[int, list[str]] = {
            index: self.get_seqs_in_pool(index) for index in range(self.n_pools)
        }

        # Adds the first valid primerpair
        for primerpair in primerpairs:
            for pool_index in range(self.n_pools):
                match self.check_primerpair_can_be_added(
                    primerpair, pool_index, pool_seqs_map[pool_index]
                ):
                    case PrimerPairCheck.OK:
                        self.add_primer_pair_to_pool(primerpair, pool_index, msa_index)
                        return SchemeReturn.ADDED_FIRST_PRIMERPAIR
                    case _:
                        continue

        # If not primerpair can be added return false
        return SchemeReturn.NO_FIRST_PRIMERPAIR

    def get_leading_coverage_edge(self) -> int:
        """This will return the furthest primer-trimmed region with coverage"""
        # This will crash if no primer has been added, but should not be called until one is
        return self._last_pp_added[-1].rprimer.start

    def get_leading_amplicon_edge(self) -> int:
        """This will return the furthest point of an amplicon"""
        # This will crash if no primer has been added, but should not be called until one is
        return max(self._last_pp_added[-1].rprimer.ends())

    def find_ol_primerpairs(self, all_pp_list, minoverlap) -> list[PrimerPair]:
        """
        Finds all primerpairs that could overlap with the last primerpair added.
        However, it does not check for clash between new PP and the last PP in the same pool
        """
        last_primer_pair: PrimerPair = self._last_pp_added[-1]
        return get_pp_window(
            all_pp_list,
            fp_end_min=last_primer_pair.fprimer.end,
            fp_end_max=last_primer_pair.rprimer.start - minoverlap,
            rp_start_min=max(last_primer_pair.rprimer.ends()) + minoverlap,
        )

    def try_ol_primerpairs(self, all_pp_list, msa_index) -> SchemeReturn:
        """
        This will try and add this primerpair into any valid pool.
        Will return true if the primerpair has been added
        """
        try:
            last_primer_pair = self._last_pp_added[-1]
            last_pool = last_primer_pair.pool
            if last_primer_pair.msa_index != msa_index:
                raise IndexError  # This will force the except block to run

        except IndexError:
            # If no primer has been added, add the first primerpair

            return self.add_first_primer_pair(all_pp_list, msa_index)

        # Find what other pools to look in
        pos_pools_indexes = [
            (last_pool + i) % self.n_pools
            for i in range(self.n_pools)
            if (last_pool + i) % self.n_pools != last_pool
        ]

        # Create a hashmap of all sequences in each pool for quick look up
        index_to_seqs: dict[int, list[str]] = {
            index: self.get_seqs_in_pool(index) for index in pos_pools_indexes
        }

        # Find pp that could ol, depending on which pool
        pos_ol_pp = self.find_ol_primerpairs(all_pp_list, self.config.min_overlap)

        # Sort the primerpairs depending on overlap score
        pos_ol_pp.sort(
            key=lambda pp: ol_pp_score(
                pp.rprimer.start,
                len(pp.all_seqs()),
                self.get_leading_coverage_edge() - self.config.min_overlap,
                self.config,
            ),
            reverse=True,
        )

        # For each primerpair
        for ol_pp in pos_ol_pp:
            # For each pool
            for pool_index in pos_pools_indexes:
                match self.check_primerpair_can_be_added(
                    ol_pp, pool_index, index_to_seqs.get(pool_index)
                ):
                    case PrimerPairCheck.OK:
                        # If the primer passes all the checks, add it to the pool
                        self.add_primer_pair_to_pool(ol_pp, pool_index, msa_index)
                        return SchemeReturn.ADDED_OL_PRIMERPAIR
                    case _:
                        continue

        # If non of the primers work, return false
        return SchemeReturn.NO_OL_PRIMERPAIR

    # backtracking
    def try_backtrack(self, all_pp_list, msa_index) -> SchemeReturn:
        """If there are no other valid ol primerpairs, replace the last primerpair added and try again"""

        # Remove the last primerpair added
        last_pp = self.remove_last_primer_pair()

        # If the last primerpair was from a different msa, add it back in and return false
        if last_pp.msa_index != msa_index:
            self.add_primer_pair_to_pool(last_pp, last_pp.pool, msa_index)
            return SchemeReturn.NO_BACKTRACK

        # Find what other pools to look in
        pos_pools_indexes = [last_pp.pool]
        # Create a hashmap of all sequences in each pool for quick look up
        index_to_seqs: dict[int, list[str]] = {
            index: self.get_seqs_in_pool(index) for index in pos_pools_indexes
        }

        is_replacement_first = False
        # If the last primerpair was the first primerpair, or first for this msa
        if (
            len(self._last_pp_added) == 0
            or self._last_pp_added[-1].msa_index != last_pp.msa_index
        ):
            # Handle adding a new first primerpair
            replacement_pps = [x for x in all_pp_list if x != last_pp]
            replacement_pps.sort(key=lambda pp: (pp.fprimer.end, -pp.rprimer.start))
            is_replacement_first = True
        else:
            # Handle adding a new ol primerpair
            # Find all primerpairs that could replace the last primerpair
            replacement_pps = [
                pp
                for pp in self.find_ol_primerpairs(all_pp_list, 1)
                if pp != last_pp  # Change minoverlap to 1 to help the solver
            ]
            # Sort the primerpair on score
            replacement_pps.sort(
                key=lambda pp: bt_ol_pp_score(
                    pp.rprimer.start,
                    len(pp.all_seqs()),
                    self.get_leading_coverage_edge() - 1,
                    self.config,
                ),
                reverse=True,
            )

        # For each replacement primerpair
        for pp in replacement_pps:
            # For each pool
            for pool_index in pos_pools_indexes:
                primers_in_same_pool = self._pools[pool_index]

                # Guard for clash between the last primer in the same pool
                # If the last primer is from the same msa and does clash, skip it
                if (
                    not is_replacement_first
                    and len(primers_in_same_pool) > 0
                    and self._pools[pool_index][-1].msa_index == msa_index
                    and max(self._pools[pool_index][-1].rprimer.ends())
                    >= min(pp.fprimer.starts())
                ):
                    continue

                # Guard for interactions
                if do_pools_interact_py(
                    pp.all_seqs(),
                    index_to_seqs.get(pool_index),
                    self.config.dimer_score,
                ):
                    continue

                # Guard for miss priming
                if detect_new_products(
                    pp.find_matches(
                        self._matchDB,
                        remove_expected=False,
                        kmersize=self.config.mismatch_kmersize,
                        fuzzy=self.config.mismatch_fuzzy,
                    ),
                    self._matches[pool_index],
                    self.config.mismatch_product_size,
                ):
                    continue

                # If all checks pass add the primerpair to the pool this is a valid alternative
                # See if the valid alternative had a valid overlap
                # Add the primer to the pool
                self.add_primer_pair_to_pool(pp, pool_index, msa_index)
                # Try and add an overlap
                match self.try_ol_primerpairs(all_pp_list, msa_index=msa_index):
                    case SchemeReturn.ADDED_OL_PRIMERPAIR:
                        # Fixed the problem
                        return SchemeReturn.ADDED_BACKTRACKED
                    case _:
                        # No overlap was found. Remove the primerpair and the try the next one
                        self.remove_last_primer_pair()
                        continue

        # If non of the primers work, add the last pp back in and return false
        self.add_primer_pair_to_pool(last_pp, last_pp.pool, msa_index)
        return SchemeReturn.NO_BACKTRACK

    def try_walk_primerpair(self, all_pp_list, msa_index) -> SchemeReturn:
        """
        Find the next valid primerpair while walking forwards
        """
        last_pool = self._last_pp_added[-1].pool
        # Find what other pools to look in, can look in same pool
        pos_pools_indexes = [
            (last_pool + i) % self.n_pools for i in range(self.n_pools)
        ]

        # Create a hashmap of all sequences in each pool for quick look up
        index_to_seqs: dict[int, list[str]] = {
            index: self.get_seqs_in_pool(index) for index in pos_pools_indexes
        }
        # Find the walking start index
        walking_min = self._last_pp_added[-1].rprimer.start - self.config.min_overlap

        # Find the first primer that could walk
        ## Use that index to slice the list
        pos_walk_pp = []
        for index, pp in enumerate(all_pp_list):
            if pp.fprimer.end > walking_min:
                pos_walk_pp = all_pp_list[index:]
                break

        # Sort walking primers by score
        pos_walk_pp.sort(
            key=lambda pp: walk_pp_score(
                pp.fprimer.end,
                len(pp.all_seqs()),
                self._last_pp_added[-1].rprimer.region()[1],
            ),
            reverse=True,
        )

        # For each primer, try each pool
        for walk_pp in pos_walk_pp:
            for pool_index in pos_pools_indexes:
                match self.check_primerpair_can_be_added(
                    walk_pp, pool_index, index_to_seqs.get(pool_index)
                ):
                    case PrimerPairCheck.OK:
                        # If the primer passes all the checks, add it to the pool
                        self.add_primer_pair_to_pool(walk_pp, pool_index, msa_index)
                        return SchemeReturn.ADDED_WALK_PRIMERPAIR
                    case _:
                        continue

        # If non of the primers work, return false
        return SchemeReturn.NO_WALK_PRIMERPAIR

    def try_circular(self, msa) -> SchemeReturn:
        """
        This will try and add a primerpair that can span from the end of the msa back to the start as if the genome was circular
        """
        try:
            first_pp: PrimerPair = self._last_pp_added[0]
            last_pp: PrimerPair = self._last_pp_added[-1]
            last_pool = last_pp.pool
        except IndexError:
            # If no primerpairs have been added
            return SchemeReturn.NO_CIRCULAR

            # Find all possible fkmers and rkmer that could span the end of the msa
        pos_fkmers = [
            fkmer
            for fkmer in msa.fkmers
            if fkmer.end < last_pp.rprimer.start
            and fkmer.end > last_pp.rprimer.start - self.config.amplicon_size_max
        ]
        pos_rkmers = [
            rkmer
            for rkmer in msa.rkmers
            if rkmer.start > first_pp.fprimer.end
            and rkmer.start < first_pp.fprimer.end + self.config.amplicon_size_max
        ]

        # Get the mapping array
        if msa._mapping_array is None:
            ref_size = len(msa.array[0])
        else:
            ref_size = max([x for x in msa._mapping_array if x is not None])

        # Check with Kmers interact
        # Create all the primerpairs
        checked_pp = []
        for fkmer in pos_fkmers:
            for rkmer in pos_rkmers:
                # Check the primerpair is the correct length
                amp_size = (ref_size - fkmer.end) + rkmer.start

                if amp_size < self.config.amplicon_size_min:
                    continue
                if amp_size > self.config.amplicon_size_max:
                    continue

                # Check for interactions
                if not which_kmers_pools_interact(
                    [fkmer], [rkmer], self.config.dimer_score, calc_all=False
                ):
                    pp = PrimerPair(fkmer, rkmer, msa.msa_index)
                    pp.chrom_name = msa._chrom_name
                    pp.amplicon_prefix = msa._uuid
                    checked_pp.append(pp)

        # Sort the primerpairs by number of primers
        checked_pp.sort(
            key=lambda pp: (
                len(pp.all_seqs()),
                (ref_size - pp.fprimer.end) + pp.rprimer.start,
            )
        )

        pos_pools_indexes = [
            (last_pool + i) % self.n_pools for i in range(self.n_pools)
        ]

        # Create a hashmap of all sequences in each pool for quick look up
        index_to_seqs: dict[int, list[str]] = {
            index: self.get_seqs_in_pool(index) for index in pos_pools_indexes
        }

        for c_pp in checked_pp:
            for pool_index in pos_pools_indexes:
                # Guard for clash between the last primer in the same pool
                if self._pools[pool_index][-1].msa_index == msa.msa_index and max(
                    self._pools[pool_index][-1].rprimer.ends()
                ) >= min(c_pp.fprimer.starts()):
                    continue

                # Guard for clash between the first primer in the same pool
                if self._pools[pool_index][0].msa_index == msa.msa_index and min(
                    self._pools[pool_index][0].fprimer.starts()
                ) <= max(c_pp.rprimer.ends()):
                    continue

                # Guard for Primer-Primer Interactions
                if do_pools_interact_py(
                    c_pp.all_seqs(),
                    index_to_seqs.get(pool_index),
                    self.config.dimer_score,
                ):
                    continue

                # Skip the miss priming product check, as we are using special indexes
                # If the primer passes all the checks, add it to the pool
                self.add_primer_pair_to_pool(c_pp, pool_index, msa.msa_index)
                return SchemeReturn.ADDED_CIRCULAR

        # No Primers could be added
        return SchemeReturn.NO_CIRCULAR

    def all_primers(self) -> list[PrimerPair]:
        all_pp = [pp for pool in (x for x in self._pools) for pp in pool]
        all_pp.sort(key=lambda pp: (str(pp.msa_index), pp.amplicon_number))
        return all_pp



# Module imports
from primaldimer_py import Kmer, do_pools_interact_py  # type: ignore

from primalscheme3.core.config import Config
from primalscheme3.core.mismatches import MatchDB
from primalscheme3.core.thermo import calc_tm, gc


class FKmer(Kmer):
    """Due to the bedfile format, the end is non-inclusive"""

    end: int
    _starts: set[int]
    _region: tuple[int, int]
    # Add slots for some performance gains
    __slots__ = ["end", "_starts", "_region"]

    def __init__(self, end, seqs) -> None:
        self.end = end
        self._starts = {self.end - x for x in self.lens()}
        self._region = (self.end - min(self.lens()), self.end)

    def region(self) -> tuple[int, int]:
        return self._region

    def len(self) -> list[int]:
        return self.lens()

    def starts(self) -> set[int]:
        return self._starts

    def __str__(self, reference, amplicon_prefix, pool) -> str:
        string_list = []
        for i, seq in enumerate(sorted(self.seqs), 1):
            string_list.append(
                f"{reference}\t{self.end-len(seq)}\t{self.end}\t{amplicon_prefix}_LEFT_{i}\t{pool}\t+\t{seq}\n"
            )
        return "".join(string_list)

    def find_matches(
        self,
        matchDB: MatchDB,
        remove_expected: bool,
        fuzzy: bool,
        kmersize: int,
        msa_index,
    ) -> set[tuple]:
        """Returns all matches of this FKmer"""
        return matchDB.find_fkmer(
            self,
            fuzzy=fuzzy,
            remove_expected=remove_expected,
            kmersize=kmersize,
            msaindex=msa_index,
        )

    def __hash__(self) -> int:
        seqs = list(self.seqs)
        seqs.sort()
        return hash(f"{self.end}{self.seqs}")

    def __eq__(self, other):
        if isinstance(other, FKmer):
            return self.__hash__() == other.__hash__()
        else:
            return False

    def remap(self, mapping_array):
        """
        Remaps the fkmer to a new indexing system
        Returns None if the fkmer is not valid
        """
        if mapping_array[self.end] is not None:
            self.end = mapping_array[self.end]
            self._starts = {self.end - len(x) for x in self.seqs}
            self._region = (self.end - min(self.lens()), self.end)
            return self
        else:
            return None


class RKmer(Kmer):
    start: int

    # Add slots for some performance gains
    __slots__ = ["start", "_ends", "_region"]
    _ends: set[int]
    _region: tuple[int, int]

    def __init__(self, start, seqs) -> None:
        self.start = start
        self._ends = {len(x) + self.start for x in self.seqs}
        self._region = (self.start, self.start + max(self.lens()))

    def region(self) -> tuple[int, int]:
        return self._region

    def len(self) -> list[int]:
        return self.lens()

    def ends(self) -> set[int]:
        return self._ends

    def __str__(self, reference, amplicon_prefix, pool) -> str:
        string_list = []
        for i, seq in enumerate(sorted(self.seqs), 1):
            string_list.append(
                f"{reference}\t{self.start}\t{self.start+len(seq)}\t{amplicon_prefix}_RIGHT_{i}\t{pool}\t-\t{seq}\n"
            )
        return "".join(string_list)

    def find_matches(
        self,
        matchDB: MatchDB,
        remove_expected: bool,
        fuzzy: bool,
        kmersize: int,
        msa_index: int,
    ) -> set[tuple]:
        """Returns all matches of this FKmer"""
        return matchDB.find_rkmer(
            self,
            fuzzy=fuzzy,
            remove_expected=remove_expected,
            kmersize=kmersize,
            msaindex=msa_index,
        )

    def __hash__(self) -> int:
        seqs = list(self.seqs)
        seqs.sort()
        return hash(f"{self.start}{self.seqs}")

    def __eq__(self, other):
        if isinstance(other, RKmer):
            return self.__hash__() == other.__hash__()
        else:
            return False

    def remap(self, mapping_array):
        """
        Remaps the rkmer to a new indexing system
        Returns None if the rkmer is not valid
        """
        if mapping_array[self.start] is not None:
            self.start = mapping_array[self.start]
            self._ends = {len(x) + self.start for x in self.seqs}
            self._region = (self.start, self.start + max(self.lens()))
            return self
        else:
            return None


class PrimerPair:
    fprimer: FKmer
    rprimer: RKmer
    amplicon_number: int
    pool: int
    msa_index: int
    chrom_name: str | None
    amplicon_prefix: str | None
    _score: float | None

    __slots__ = [
        "fprimer",
        "rprimer",
        "amplicon_number",
        "pool",
        "msa_index",
        "chrom_name",
        "amplicon_prefix",
        "_score",
    ]

    def __init__(
        self,
        fprimer,
        rprimer,
        msa_index,
        amplicon_number=-1,
        pool=-1,
    ):
        self.fprimer = fprimer
        self.rprimer = rprimer
        self.amplicon_number = amplicon_number
        self.pool = pool
        self.msa_index = msa_index
        self.chrom_name = None
        self.amplicon_prefix = None
        self._score = None

    def get_score(self, target_gc=0.5):
        """
        Returns the mean gc diff of the primerpair
        """
        if self._score is None:
            self._score = sum(
                [abs(target_gc - (gc(x) / 100)) for x in self.all_seqs()]
            ) / len(self.all_seqs())
        return self._score

    def regions(self) -> tuple[tuple[int, int], tuple[int, int]]:
        return self.fprimer.region(), self.rprimer.region()

    def set_amplicon_number(self, amplicon_number) -> None:
        self.amplicon_number = amplicon_number

    def set_pool_number(self, pool_number) -> None:
        self.amplicon_number = pool_number

    def find_matches(self, matchDB, fuzzy, remove_expected, kmersize) -> set[tuple]:
        """
        Find matches for the FKmer and RKmer
        """
        matches = set()
        # Find the FKmer matches
        matches.update(
            self.fprimer.find_matches(
                matchDB, fuzzy, remove_expected, kmersize, msa_index=self.msa_index
            )
        )
        # Find the RKmer matches
        matches.update(
            self.rprimer.find_matches(
                matchDB, fuzzy, remove_expected, kmersize, self.msa_index
            )
        )
        return matches

    def kmers(self):
        """
        Returns the FKmer and RKmer
        """
        return self.fprimer, self.rprimer

    def primertrimmed_region(self) -> tuple[int, int]:
        """
        Returns the region of the primertrimed region
        Right position is non-inclusive
        """
        return self.fprimer.end, self.rprimer.start

    def inter_free(self, cfg) -> bool:
        """
        True means interaction
        """
        return do_pools_interact_py(
            [*self.fprimer.seqs], [*self.rprimer.seqs], cfg["dimerscore"]
        )

    def all_seqs(self) -> list[str]:
        return [x for x in self.fprimer.seqs] + [x for x in self.rprimer.seqs]

    def calc_tm(self, config: Config) -> list[float]:
        """
        Calculates the tm for all primers in the PrimerPair
        :param cfg: config dict
        :return: list of tm values
        """
        return [
            calc_tm(
                seq,
                mv_conc=config.mv_conc,
                dv_conc=config.dv_conc,
                dna_conc=config.dna_conc,
                dntp_conc=config.dna_conc,
            )
            for seq in self.all_seqs()
        ]

    def __hash__(self) -> int:
        return hash(f"{self.regions()}{self.all_seqs()}")

    def __eq__(self, other):
        if isinstance(other, PrimerPair):
            return self.__hash__() == other.__hash__()
        else:
            return False

    def to_bed(self) -> str:
        """
        Turns the primerpair into a string for a bed file
        :param chromname: name of the chromosome
        :param amplicon_prefix: prefix for the amplicon
        :return: string for the bed file
        """
        return self.__str__()

    def __str__(self):
        return self.fprimer.__str__(
            reference=f"{self.chrom_name}",
            amplicon_prefix=f"{self.amplicon_prefix}_{self.amplicon_number}",
            pool=self.pool + 1,
        ) + self.rprimer.__str__(
            reference=f"{self.chrom_name}",
            amplicon_prefix=f"{self.amplicon_prefix}_{self.amplicon_number}",
            pool=self.pool + 1,
        )



"""
These are binary searches implementions
"""


def get_r_window_FAST2(kmers, start: int, end: int):
    """
    This will perform a binary search on the list of kmers.
    The Kmer start positions will be used as the search value
    """
    included_kmers = []
    n_kmers = len(kmers)
    high = n_kmers - 1
    low = 0
    mid = 0
    while low <= high:
        mid = (high + low) // 2
        # If the midpoint is inside the range
        if start <= kmers[mid].start <= end:
            while True:
                # Walk back until first value, or the first position
                if mid == 0:
                    break
                elif kmers[mid - 1].start >= start:
                    mid -= 1
                else:
                    break
            # Mid is now the first value so walk forwards
            while True:
                if mid < n_kmers and kmers[mid].start <= end:
                    included_kmers.append(kmers[mid])
                    mid += 1
                else:
                    return included_kmers
        # If start is greater ignore the left half
        elif kmers[mid].start < start:
            low = mid + 1
        # If start is smaller ignore the right half
        elif kmers[mid].start > end:
            high = mid - 1

    # If the code reaches here there are no KMERS within the list inside the range
    ## Return an empty list for continuity
    return []

    # These are modifed get_window_fast for the new kmers classes


def get_f_window_FAST2(kmers, start: int, end: int):
    """
    This will perform a binary search on the list of kmers.
    The Kmer end position will be used as the search value
    """
    included_kmers = []
    n_kmers = len(kmers)
    high = n_kmers - 1
    low = 0
    mid = 0
    while low <= high:
        mid = (high + low) // 2
        # If the midpoint is inside the range
        if start <= kmers[mid].end <= end:
            while True:
                # Walk back until first value, or the first position
                if mid == 0:
                    break
                elif kmers[mid - 1].end >= start:
                    mid -= 1
                else:
                    break
            # Mid is now the first value so walk forwards
            while True:
                if mid < n_kmers and kmers[mid].end <= end:
                    included_kmers.append(kmers[mid])
                    mid += 1
                else:
                    return included_kmers
        # If start is greater ignore the left half
        elif kmers[mid].end < start:
            low = mid + 1
        # If start is smaller ignore the right half
        elif kmers[mid].end > end:
            high = mid - 1

    # If the code reaches here there are no KMERS within the list inside the range
    ## Return an empty list for continuity
    return []


def get_pp_window(pp, fp_end_min: int, fp_end_max: int, rp_start_min: int):
    """
    This will perform a semi-binary search on the list of primerpairs.
    The primerpair.fprimer.end position will be used as the search value
    """
    # If you want to add a PP to Pool 1

    #                            fp_end_min
    #                            |       fp_end_max
    # Pool 1    XXX------------XXX       |
    # Pool 2              XXX------------XXX
    #                                      |
    #                                      rp_start_min
    included_pp = []
    n_pp = len(pp)
    high = n_pp - 1
    low = 0
    mid = 0
    while low <= high:
        mid = (high + low) // 2
        # If the midpoint is inside the range
        if fp_end_min <= pp[mid].fprimer.end <= fp_end_max:
            while True:
                # Walk back until first value, or the first position
                if mid == 0:
                    break
                elif pp[mid - 1].fprimer.end >= fp_end_min:
                    mid -= 1
                else:
                    break
            # Mid is now the first value so walk forwards
            while mid < n_pp and pp[mid].fprimer.end <= fp_end_max:
                # If the rprimer.start is
                if pp[mid].rprimer.start >= rp_start_min:
                    included_pp.append(pp[mid])

                mid += 1
            return included_pp
        # If start is greater ignore the left half
        elif pp[mid].fprimer.end < fp_end_min:
            low = mid + 1
        # If start is smaller ignore the right half
        elif pp[mid].fprimer.end > fp_end_max:
            high = mid - 1

    # If the code reaches here there are no KMERS within the list inside the range
    ## Return an empty list for continuity
    return []

    # These are modifed get_window_fast for the new kmers classes
